# -*- coding: utf-8 -*-
'''
Copyright 2012 Rodrigo Pinheiro Matias <rodrigopmatias@gmail.com>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''
from pygooglechart import *

try:
    from PIL import Image
    from PIL import ImageDraw
    from PIL import ImageColor
    from PIL import ImageFont
except ImportError:
    import Image
    import ImageDraw
    import ImageColor
    import ImageFont


import os
import re
import sys
import signal
import time
import sqlite3 as sqlite
import math

try:
    font_dir = os.path.abspath('%s/../data' % os.path.dirname(os.path.abspath(__file__)))

    font = ImageFont.truetype(os.path.join(font_dir, 'UbuntuMono-R.ttf'), 11)
    font_title = ImageFont.truetype(os.path.join(font_dir, 'UbuntuMono-B.ttf'), 18)
except:
    font = None
    font_title = None

DEBUG = False
# DEBUG = True

class BasePlugin(object):

    __connector = None

    def _get_database_path(self):
        return os.path.expanduser('~/.hwm.db')

    def get_connector(self):
        return sqlite.connect(self._get_database_path())

    def __init__(self, params):
        self._params = params
        self._die = False
        self._period = 300.0
    
    @classmethod
    def description(class_):
        return {
            'name': None,
            'slug': None,
            'options': []
        }

class BaseColectorPlugin(BasePlugin):

    def clear(self):
        raise NotImplementedError('Method abstract')

    def _prepare(self, **kargs):
        raise NotImplementedError('Method abstract')

    def get_data_sequence(self, **kargs):
        return self._get_data_sequencce(**kargs)

    def get_column(self):
        raise NotImplementedError('Method abstract')

    def reduce_number(self, number):
        if number == 0:
            return 0, 1
        else:
            return float(number) / (float(10) ** int(math.log(number, 10))), math.log(number, 10)

    def calc_limits(self, start = None, end = None):
        if end is None: raise Exception('Deve especificar o final do range')
        if start is None: start = 0

        base, exp = self.reduce_number(end)
        return start, int((int(base) + 1) * (10 ** int(exp)))

    def _get_data_sequencce(self, **kargs):
        seq = [seq for seq in self.do_extract(**kargs)]
        return {
            'column': self.get_column(),
            'collection': tuple(seq),
            'count': len(seq)
        }

    def _create_table(self):
        raise NotImplementedError('Method abstract')
    
    def do_colect(self):
        raise NotImplementedError('Method abstract')
    
    def do_extract(self, **kargs):
        raise NotImplementedError('Method abstract')

    def handle_sigint(self, signal, frame):
        self._die = True
        print ''
        print 'Processo finalizado'
        print ''

    def colect(self):
        self.do_colect()

    def extract(self):
        params = self._params.__dict__
        self.do_extract(**params)

class BaseGraphicPlugin(BasePlugin):

    @property
    def params(self):
        raise NotImplementedError('Property is abstract')

    @property
    def filename(self):
        raise NotImplementedError('Property is abstract')

    @property
    def colors(self):
        raise NotImplementedError('Property is abstract')

    def _config_markers(self, chart, index, color, fill = True):
        if fill is True:
            chart.add_marker(index, 0, 'B', '%s90' % color, 2)
        else:
            chart.add_marker(index, 0, 'B', '%s40' % color, 2)

        chart.add_marker(index, -1, 'o', color, 5)
        chart.add_marker(index, -1, 'o', 'ffffff', 3)

    def humanize(self, number):
        return (number, '')

    def _legend(self, canvas):
        legends = dict(zip(self.get_column(), self.colors))
        x, y = 10, 200
        for text, color in legends.items():
            canvas.rectangle(
                ((x, y), (x + 12, y + 12)),
                ** {
                    'fill': ImageColor.getcolor('#%s' % color, 'RGB'),
                    'outline': ImageColor.getcolor('#%s' % color, 'RGB')
                }
            )
            canvas.text(
                (x + 15, y), 
                text,
                ** {
                    'fill': ImageColor.getcolor('#000', 'RGB'),
                    'font': font
                }
            )
            y += 15

    def _title(self, canvas):
        title = self.params.get('label')
        width, height = canvas.textsize(title, **{'font': font_title})

        x = int(canvas.im.size[0] / 2.0 - width / 2.0)
        y = 10

        canvas.text((x, y), title, **{
            'fill': ImageColor.getcolor('#000', 'RGB'),
            'font': font_title
        })

    def _framing(self, filename):
        grap = Image.open(filename)
        frame = Image.new('RGBA', (675, 200 + (15 * len(self.get_column()))), ImageColor.getcolor('#fff', 'RGB'))
        canvas = ImageDraw.Draw(frame)

        canvas.rectangle(
            ((5,35), (670, 195)), 
            **{
                'fill': ImageColor.getcolor('#fff', 'RGB'),
                'outline': ImageColor.getcolor('#aaa', 'RGB')
            }
        )

        self._legend(canvas)
        self._title(canvas)

        frame.paste(grap, (12, 40))

        if DEBUG is True:
            frame.show()
        else:
            frame.save(filename,  'PNG')

    def draw(self):
        data = BaseColectorPlugin.get_data_sequence(self, **self.params)
        filename = self.filename

        max_y = 0
        for item in data.get('collection')[1:]:
            for value in item: max_y = value if value > max_y else max_y

        min_y, max_y = self.calc_limits(end = max_y)
        space_y = max_y - min_y
        part_y = (space_y / 10)

        slot_count = len(data.get('collection')[0])

        chart = SimpleLineChart(650, 150, y_range = (min_y, max_y), colours = self.colors)
        for slot in xrange(slot_count):
            chart.add_data([item[slot] for item in data.get('collection')[1:]])
            self._config_markers(chart, slot, self.colors[slot])
            
        chart.set_axis_labels(Axis.LEFT, ['%0.1f %s' % self.humanize(x) for x in xrange(min_y, max_y, part_y)])
        chart.set_axis_labels(Axis.RIGHT, ['%0.1f %s' % self.humanize(x) for x in xrange(min_y, max_y, part_y)])
        chart.set_grid(
            50,
            11.15,
            1,
            2
        )

        chart.download(filename)
        self._framing(filename)

class PluginManager:

    class PluginNotFound(Exception):

        def __init__(self, slug):
            Exception.__init__(self, u'The slug %s not found in base of plugins.' % slug)

    class BasePluginNotCompatible(Exception):

        def __init__(self):
            Exception.__init__(self, u'The base not is an BasePlugin valid.')


    def __init__(self, base):
        if issubclass(base, BasePlugin) is True:
            self._database = {}
            self._basePlugin = base
            self._scan()
        else:
            raise self.BasePluginNotCompatible()

    def _scan(self):
        pdir = os.path.dirname(os.path.abspath(__file__))
        to_match = '^(?P<module>.+)\.py$'
        files = [filename for filename in os.listdir(pdir) if re.match(to_match, filename)]
        modules = [re.match(to_match, filename).groupdict().get('module') for filename in files if filename not in ['__init__.py', 'base.py']]

        for module in modules:
            sys.stdout.write('Loading module %s ... ' % module)
            loaded = None
            try:
                exec('from hwm.plugin import %s as loaded' % module)
            except:
                print 'fail'
            else:
                in_module = []
                for attr in dir(loaded):
                    flag = False
                    plg = getattr(loaded, attr)
                    try:
                        flag = (attr != 'BaseColectorPlugin' and issubclass(plg, self._basePlugin))
                    except:
                        pass
                    else:
                        if flag is True:
                            self._database.update({plg.description().get('slug'): plg})
                            in_module.append(str(plg.description().get('name')))
                sys.stdout.write('(')
                sys.stdout.write(', '.join(in_module))
                sys.stdout.write(')')

                print ' ok'

    def get(self, slug):
        if slug in self._database.keys():
            return self._database.get(slug)
        else:
            raise self.PluginNotFound(slug)